/***************************************************************************

 * Title:          stdinc.h

 * Author:         Haixu Tang

 * Created:        Jun. 2002

 * Last modified:  May. 2004

 *

 * Copyright (c) 2001-2004 The Regents of the University of California

 * All Rights Reserved

 * See file LICENSE for details.

 ***************************************************************************/

#include <limits.h>

#include <stdio.h>

#include <math.h>

#include <unistd.h>

#include <string.h>

#include <stdlib.h>

#include <time.h>

#include <pthread.h>

#include "def.h"

#include "general.h"

