#ifndef __TYPES_RJ
#define __TYPES_RJ

typedef unsigned long long	ubyte8;
typedef unsigned int		ubyte4;
typedef unsigned short		ubyte2;
typedef unsigned char		ubyte;

typedef long long		byte8;
typedef int 			byte4;
typedef short			byte2;
typedef char			byte;

#endif
