#ifndef KSTRING_H
#define KSTRING_H

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#ifndef kroundup32
#define kroundup32(x) (--(x), (x)|=(x)>>1, (x)|=(x)>>2, (x)|=(x)>>4, (x)|=(x)>>8, (x)|=(x)>>16, ++(x))
#endif

#ifndef KSTRING_T
#define KSTRING_T kstring_t
typedef struct __kstring_t
{
  size_t l, m;
  char *s;
} kstring_t;
#endif

int ksprintf ( kstring_t *s, const char *fmt, ... );
int ksplit_core ( char *s, int delimiter, int *_max, int **_offsets );

// calculate the auxiliary array, allocated by calloc()
int *ksBM_prep ( const uint8_t *pat, int m );

/* Search pat in str and returned the list of matches. The size of the
 * list is returned as n_matches. _prep is the array returned by
 * ksBM_prep(). If it is a NULL pointer, ksBM_prep() will be called. */
int *ksBM_search ( const uint8_t *str, int n, const uint8_t *pat, int m, int *_prep, int *n_matches );

static inline int kputsn ( const char *p, int l, kstring_t *s )
{
  if ( s->l + l + 1 >= s->m )
    {
      s->m = s->l + l + 2;
      kroundup32 ( s->m );
      s->s = ( char * ) realloc ( s->s, s->m );
    }

  strncpy ( s->s + s->l, p, l );
  s->l += l;
  s->s[s->l] = 0;
  return l;
}

static inline int kputs ( const char *p, kstring_t *s )
{
  return kputsn ( p, strlen ( p ), s );
}

static inline int kputc ( int c, kstring_t *s )
{
  if ( s->l + 1 >= s->m )
    {
      s->m = s->l + 2;
      kroundup32 ( s->m );
      s->s = ( char * ) realloc ( s->s, s->m );
    }

  s->s[s->l++] = c;
  s->s[s->l] = 0;
  return c;
}

static inline int *ksplit ( kstring_t *s, int delimiter, int *n )
{
  int max = 0, *offsets = 0;
  *n = ksplit_core ( s->s, delimiter, &max, &offsets );
  return offsets;
}

#endif
